{-
  Repraesenation eines Sets unter Verwendung von Listen
  Quelle: Richard Bird - Introduction to Functional Programming using Haskell
-}
                      
module LSet(LSet, empty, isEmpty, insert, member, delete)
where
  newtype LSet a = LSt ([a])

  --Dictionary-Funktionen
  
  --leere Menge
  empty :: LSet a
  empty = LSt []
  
  --TeLSt auf leere Menge
  isEmpty :: LSet a -> Bool
  isEmpty (LSt x) = null x
    
  --verbesserte Member-Funktion, arbeitet auf Basis einer sortierten Liste
  member :: (Ord a) => LSet a -> a -> Bool
  member (LSt xs) x = if null ys then False
                          else (x == head ys)
		                      where
                            ys = dropWhile (<x) $ xs
  
  --loescht ein Element aus der Menge
  delete :: (Eq a) => a -> LSet a -> LSet a
  delete x (LSt xs) = LSt(filter (/=x) xs)

  --insert-Funktion ohne Duplikate, sortiert
  insert :: (Ord a) => a -> LSet a -> LSet a
  insert x (LSt xs) = LSt(ys ++ [x] ++ (filter (/=x) zs))
                             where
                             (ys, zs) = span (<x) xs  

  --Ausgabe eines Sets in der ueblichen Form {[Werte]}
  instance Show a => Show (LSet a)
    where
    show (LSt xs)
      |null xs = "{}"
      |otherwise = '{':show' (LSt xs) ++ "}"
      where
        show' (LSt([])) = "" 
        show' (LSt(x:xs)) = show x ++ " " ++ show' (LSt xs)
        

  --weitere Funktionen // unterschiedliche

  
    --Hilfsfunktionen
  some :: (a -> Bool) -> LSet a -> Bool
  some p xs = or . map p $ reprn xs
  
  abLStr :: (Ord a) => [a] -> LSet a
  abLStr = foldr insert empty
  

  reprn :: LSet a -> [a]
  reprn (LSt []) = []
  reprn (LSt(x:xs))
    |null xs   = [x]
    |otherwise = [x] ++ reprn (LSt xs)


  --Test, ob Element in Menge iLSt
  member' :: (Eq a) => LSet a -> a -> Bool
  member' xs x = some (==x) xs  
  
  --fuegt ein Element der Menge hinzu
  insert' :: a -> LSet a -> LSet a
  insert' x (LSt xs) = LSt(x:xs)
  
  --verbesserte insert-Funktion fuer Menge ohne Duplikate
  insert'' :: (Eq a) => a -> LSet a -> LSet a
  insert'' x (LSt xs) = LSt(x:(filter (/=x) xs))

  --Vereinigung von zwei Mengen
  union :: LSet a -> LSet a -> LSet a
  union (LSt xs) (LSt ys) = LSt(xs ++ ys)
 
    --verbesserte union-Funktion fuer Menge ohne Duplikate
  union' :: (Eq a) => LSet a -> LSet a -> LSet a
  union' (LSt xs) (LSt ys) = LSt(xs ++ filter (not . member' (LSt xs)) ys)

  --noch bessere union-Funktion
  union'' :: (Ord a) => LSet a -> LSet a -> LSet a
  union'' (LSt []) ys = ys
  union'' xs (LSt []) = xs
  union'' (LSt(x:xs)) (LSt(y:ys))
    |(x < y)  = LSt (x:reprn (union'' (LSt xs) (LSt(y:ys))))
    |(x == y) = LSt (x:reprn (union'' (LSt xs) (LSt ys)))
    |(x > y)  = LSt (y:reprn (union'' (LSt (x:xs)) (LSt ys)))

  --Differenz von zwei Mengen
  minus :: (Eq a) => LSet a -> LSet a -> LSet a
  minus (LSt xs) (LSt ys) = LSt(filter (not . member' (LSt ys)) xs)
  
  minus' :: (Ord a) => LSet a -> LSet a -> LSet a
  minus' xs (LSt []) = xs
  minus' (LSt []) ys = LSt [] 
  minus' (LSt (x:xs)) (LSt (y:ys))        
    |(x < y) = LSt (x:reprn(minus' (LSt xs) (LSt (y:ys))))
    |(x == y) = minus' (LSt xs) (LSt ys)
    |(x > y) =  minus' (LSt (x:xs)) (LSt ys)

  --Schnittmenge
  meet :: (Ord a) => LSet a -> LSet a -> LSet a
  meet xs (LSt [])    = empty
  meet xs (LSt(y:ys)) = if member' xs y then insert y (meet xs (LSt ys))
                           else meet xs (LSt ys)
   
  meet' :: (Ord a) => LSet a -> LSet a -> LSet a
  meet' (LSt []) ys = LSt [] 
  meet' xs (LSt []) = LSt []
  meet' (LSt (x:xs)) (LSt (y:ys))
    |(x < y)  = meet' (LSt xs) (LSt(y:ys))
    |(x == y) = LSt (x:reprn (meet' (LSt xs) (LSt ys)))
    |(x > y)  = meet' (LSt (x:xs)) (LSt ys)   
  
